#ifndef __GPNode_H__
#define __GPNode_H__

#include <string>
#include <vector>
#include <iostream>
#include <stdexcept>
using namespace std;

#include "Function.h"
#include "rand-util.h"
#include "auto_list.h"
#include "memory_pool.h"

template<class T> class GPNode {

public:
	explicit GPNode( const GPNode<T>* parent );
	~GPNode();
	
public:
	void execute( T& output ); 
	int depth() const;
	int tree_size() const;
	int tree_depth() const;

	int num_children() const { return function_->arity(); };
	      GPNode& child( int i ) { return children_[ i ]; };
	const GPNode& child( int i ) const { return children_[ i ]; };

	bool contains_function( Function<T>* function ) const;

	void set_function( Function<T>* function ) { assert( function->arity() == function_->arity() ); function_ = function; };

	void initialise( int depth, auto_list< Function<T> >& terminals, auto_list< Function<T> >& functions, bool full );

	void print( ostream& output, int indent ) const;
	void read( istream& input, auto_list< Function<T> >& terminals, auto_list< Function<T> >& functions );

public:
	void operator=( const GPNode<T>& rhs );
	bool operator==( const GPNode<T>& rhs ) const;

private:
	void delete_children();

protected:	
	const GPNode<T>* parent_;
	auto_list< GPNode<T> > children_;
	Function<T>* function_;
	
public:
	static void* operator new(size_t size);
	static void operator delete(void* ptr);
protected:
	static memory_pool memory_pool_;
	
};

template<class T> memory_pool GPNode<T>::memory_pool_;

template<class T> void* GPNode<T>::operator new(size_t size) {
	// cout << "new" << endl;
	if ( memory_pool_.empty() )
		return ::operator new(size);
	return memory_pool_.pop();
}

template<class T> void GPNode<T>::operator delete(void* ptr) {
	//cout << "delete" << endl;
	memory_pool_.push( ptr );
}

template<class T> GPNode<T>::GPNode( const GPNode<T>* parent )
 : parent_( parent ),
   function_( NULL )
{
	children_.clear();
}

template<class T> GPNode<T>::~GPNode() {
	parent_ = NULL;
	delete_children();
	function_ = NULL;
}

template<class T> void GPNode<T>::delete_children() {
	children_.clear();
}

template<class T> void GPNode<T>::execute( T& output ) {
	assert( function_ != NULL );
	auto_list<T> inputs;
	for ( int i = 0; i < function_->arity(); i++ ) {
		inputs.add( new T() );
		children_[ i ].execute( inputs[ i ] );
	}
	function_->execute( inputs, output );
}

template<class T> bool GPNode<T>::contains_function( Function<T>* function ) const {
	if ( function_ == function ) 
		return true;
	for ( int i = 0; i < num_children(); i++ ) {
		if ( child( i ).contains_function( function ) )
			return true;
	}
	return false;
}

template<class T> int GPNode<T>::depth() const {
	return parent_ == NULL? 0 : 1 + parent_->depth();
}

template<class T> int GPNode<T>::tree_size() const {
	int size = 1;
	for ( int i = 0; i < num_children(); i++ ) {
		size += child( i ).tree_size();
	}
	return size;
}

template<class T> int GPNode<T>::tree_depth() const {
	int maxdepth = 0;
	for ( int i = 0; i < num_children(); i++ ) {
		int depth = child( i ).tree_depth()+1;
		if ( depth > maxdepth ) maxdepth = depth;
	}
	return maxdepth;
}


template<class T> void GPNode<T>::initialise( int depth, auto_list< Function<T> >& terminals, auto_list< Function<T> >& functions, bool full ) {
	delete_children();

	assert( terminals.size() != 0 );

	bool choosefunction = full && !functions.empty();
	if ( !full ) {
		unsigned int i = randrange( functions.size() + terminals.size() );
		choosefunction = (i < functions.size());
	}
	
	if ( this->depth() < depth  && choosefunction ) {
		// pick a function
		int i = randrange( functions.size() );
		function_ = &functions[ i ];
		for ( int i = 0; i < function_->arity(); i++ ) {
			children_.add( new GPNode<T>( this ) );
			children_[ i ].initialise( depth, terminals, functions, full );
		}
	}
	else {
		// pick a terminal
		int i = randrange( terminals.size() );
		function_ = &terminals[ i ];
	}
}

template<class T> void GPNode<T>::print( ostream& output, int indent ) const {
	assert( function_ != NULL );
	for ( int i = 0; i < indent; i++ ) output << '\t';
	if ( function_->arity() != 0 ) output << "( ";
	output << function_->name() << endl;
	for ( int i = 0; i < function_->arity(); i++ ) {
		const GPNode<T>& child = children_[ i ];
		child.print( output, indent+1 );
	}
	if ( function_->arity() != 0 ) {
		for ( int i = 0; i < indent; i++ ) output << '\t';
		output << ")";
		output << endl;
	}
	
}

template<class T> void GPNode<T>::read( istream& input, auto_list< Function<T> >& terminals, auto_list< Function<T> >& functions ) {
	delete_children();
	
	string token;
	string name;
	
	input >> token;
	
	if ( token == "(" )
		input >> name;
	else
		name = token; // it's probably a terminal
		
	bool found = false;
	for ( unsigned int i = 0; i < terminals.size(); i++ ) {
		Function<T>* terminal = &terminals[ i ];
		if ( name == terminal->name() ) {
			found = true;
			function_ = terminal;
			break;
		}
	}
	
	if ( !found ) {
	
		for ( unsigned int i = 0; i < functions.size(); i++ ) {
			Function<T>* function = &functions[ i ];
			if ( name == function->name() ) {
				found = true;
				function_ = function;
				for ( int i = 0; i < function_->arity(); i++ ) {
					children_.add( new GPNode<T>( this ) );
					GPNode<T>& node = children_[ i ];
					node.read( input, terminals, functions );
				}
				break;
			}
		}
		
	}
	
	if ( token == "(" ) {
		// must close the bracket if we have used one
		input >> token;
		if ( token != ")" )
			throw runtime_error( "Bracket not closed" );
	}
	
	if ( !found ) {
		cout << name << " is not a known function or terminal" << endl;
		throw runtime_error( "Unkown function/terminal" );
	}
		
}

template<class T> void GPNode<T>::operator=( const GPNode<T>& rhs ) {
	if( this == &rhs )
		return;
		
	delete_children();
	function_ = rhs.function_;
	for ( int i = 0; i < function_->arity(); i++ ) {
		children_.add( new GPNode<T>( this ) );
		GPNode<T>& node = children_[ i ];
		node = rhs.children_[ i ];
	}

}

template<class T> bool GPNode<T>::operator==( const GPNode<T>& rhs ) const {
	if( this == &rhs )
		return true;
	
	if ( function_ != rhs.function_ )
		return false;
		
	for ( int i = 0; i < function_->arity(); i++ ) {
		if ( !(children_[ i ] == rhs.children_[ i ]) ) {
			return false;
		}
	}
	return true;
}


template<class T> ostream& operator<<(ostream& output, const GPNode<T>& node ) {
    node.print( output, 0 );
    return output;
}

/*template<class T> istream& operator>>(istream& input, GPNode<T>& node ) {
    node.read( input );
    return input;
}*/


#endif // __GPNode_H__
