#ifndef __GPTree_H__
#define __GPTree_H__

#include <iostream>
#include <vector>
using namespace std;

#include "GPNode.h"
#include "auto_list.h"

template<class T> class GPTree {

public:
	GPTree( int depth, vector<T*> inputs, auto_list< Function<T> >& terminals, auto_list< Function<T> >& functions );
	virtual ~GPTree();
	
public:
	void initialise( int depth, bool full = true );
	void evaluate( T& output ) { root_.execute( output ); };
	void mutate_subtree();
	void mutate_terminal();
	void mutate_root();
	void mutate_function();
	void recombine( GPTree<T>& partner );
	void print( ostream& output ) const { output << root_; };
	void read( istream& input );

	int size() const;
	vector<T*> inputs() const { return inputs_; };
	
	bool contains_function( Function<T>* function ) const;

	GPNode<T>& randomNode();

public:
	void operator=( const GPTree<T>& rhs );
	bool operator==( const GPTree<T>& rhs ) const;
	
protected:
	const int depth_;
	GPNode<T> root_;
	vector<T*> inputs_;
	auto_list< Function<T> >& terminals_;
	auto_list< Function<T> >& functions_;
};

template<class T> GPTree<T>::GPTree( int depth, vector<T*> inputs, auto_list< Function<T> >& terminals, auto_list< Function<T> >& functions )
 : depth_( depth ),
   root_( NULL ),
   inputs_( inputs ),
   terminals_( terminals ),
   functions_( functions )
{

}

template<class T> GPTree<T>::~GPTree() {

}

template<class T> void GPTree<T>::initialise( int depth, bool full ) {
	root_.initialise( depth, terminals_, functions_, full );
}

template<class T> GPNode<T>& GPTree<T>::randomNode() {
	// random walk through the tree
	vector< GPNode<T>* > nodes;
	
	GPNode<T>* current = &root_;
	
	do {
		nodes.push_back( current );
		if ( current->num_children() == 0 )
			current = NULL;
		else {
			current = &current->child( randrange( current->num_children() ) );
		}
	}
	while( current != NULL );
	
	GPNode<T>* node = nodes[ randrange( nodes.size() ) ];

	return *node;
}

template<class T> void GPTree<T>::mutate_subtree() {
	GPNode<T>& node = randomNode();
	int node_depth = node.depth();
	int depth = node_depth + randrange( depth_ - node_depth ) +1;
	depth = depth > depth_? depth_ : depth;
	
	bool full = randdouble() < 0.5;
	node.initialise( depth, terminals_, functions_, full );
}

template<class T> void GPTree<T>::mutate_terminal() {
	GPNode<T>& node = randomNode();
	
	// effectively removes the subtree
	auto_list< Function<T> > functions;
	node.initialise( depth_, terminals_, functions, true );
}

template<class T> void GPTree<T>::mutate_root() {
	GPNode<T>& node = randomNode();
	
	if ( node.num_children() > 0 && randdouble() < 0.5 ) {
		// turn the parent into one of it's children
		GPNode<T>& child = node.child( randrange( node.num_children() ) );
		GPNode<T> tmp( NULL );
		tmp = child;
		node = tmp;
	}
	else {
		// insert a new parent above this node
		GPNode<T> tmp( NULL );
		tmp = node;
		int node_depth = node.depth();
		int depth = node_depth + randrange( depth_ - node_depth ) +1;
		depth = depth > depth_? depth_ : depth;
	
		bool full = randdouble() < 0.5;
		node.initialise( depth, terminals_, functions_, full );
		if ( node.num_children() > 0 ) {
			// copy the node to one of the new children
			node.child( randrange( node.num_children() ) ) = tmp;
		}
	}
}

template<class T> void GPTree<T>::mutate_function() {
	GPNode<T>& node = randomNode();
	Function<T>* function = NULL;
	if ( node.num_children() == 0 ) {
		function = &terminals_[ randrange( terminals_.size() ) ];
	}
	else { 
		vector< Function<T>* > matchingFunctions;
		for ( unsigned int i = 0; i < functions_.size(); i++ ) {
			Function<T>& fn = functions_[ i ];
			if ( fn.arity() == node.num_children() ) {
				matchingFunctions.push_back( &fn );
			}
		}
		assert( matchingFunctions.size() != 0 );
		function = matchingFunctions[ randrange( matchingFunctions.size() ) ];
	}
	if ( function != NULL )
		node.set_function( function );
}

template<class T> void GPTree<T>::recombine( GPTree<T>& partner ) {
	if ( this == &partner )
		return;
		
	GPNode<T>& site1 = randomNode();
	int site1depth   = site1.depth();
	// try a few times to pick a node with the same depth
	for ( int i = 0; i < 5; i++ ) {
		GPNode<T>& site2 = partner.randomNode();
	
		// only allow recombination that does not take us over depth limit
		if ( (site1depth + site2.tree_depth()) <= depth_ ) {
			site1 = site2;
			break;
		}
	}
}

template<class T> int GPTree<T>::size() const {
	return root_.tree_size();
}

template<class T> bool GPTree<T>::contains_function( Function<T>* function ) const {
	return root_.contains_function( function );
}

template<class T> void GPTree<T>::read( istream& input ) {
	root_.read( input, terminals_, functions_ );
}

template<class T> void GPTree<T>::operator=( const GPTree<T>& rhs ) {
	if ( this == &rhs )
		return;
	assert( depth_ == rhs.depth_ );
	root_      = rhs.root_;
	terminals_ = rhs.terminals_;
	functions_ = rhs.functions_;
}

template<class T> bool GPTree<T>::operator==( const GPTree<T>& rhs ) const {
	if ( this == &rhs )
		return true;
	if ( depth_ != rhs.depth_ )
		return false;
		
	return root_ == rhs.root_;
}


template<class T> ostream& operator<<(ostream& output, const GPTree<T>& tree ) {
	tree.print( output );
	return output;
}

template<class T> istream& operator>>(istream& input, GPTree<T>& tree ) {
	tree.read( input );
	return input;
}


#endif // __GPTree_H__
