#include "Image.h"

#include "rand-util.h"

#include "opencv/cv.h"
#include "opencv/highgui.h"

#include <iostream>
#include <stdexcept>
using namespace std;

unsigned int Image::DEPTH = IPL_DEPTH_32F;

Image::Image()
 : image_( NULL ),
   w_( -1 ),
   h_( -1 )
{

}

Image::Image( int w, int h, int depth )
 : image_( NULL ),
   w_( w ),
   h_( h )
{
	CvSize imgSize;
	imgSize.width = w;
	imgSize.height = h;
	image_ = cvCreateImage( imgSize, depth, 1);
}
	
Image::Image( const Image& img )
 : image_( NULL ),
   w_( -1 ),
   h_( -1 )
{
	*this = img;
}

Image::Image( const char* filename )
 : image_( NULL ),
   w_( -1 ),
   h_( -1 )
{
	load( filename );
}
	
Image::~Image() {
	//cout << "~Image() begin" << endl;
	cvReleaseImage( &image_ );
	//cout << "~Image() end" << endl;
}
		
void Image::load( const char* filename ) {
	cvReleaseImage( &image_ );
	IplImage* image = cvLoadImage( filename, 0 ); // loads as black and white
	CvSize size;
	size = cvGetSize( image );
	image_ = cvCreateImage( size, DEPTH, 1);
	w_ = size.width;
	h_ = size.height;
	cvConvertScale( image, image_, 1.0/255.0 );
	cvReleaseImage( &image );
}

void Image::flood( double value, int x, int y, double lo, double up ) {
	CvPoint seed;
	seed.x = x; 
	seed.y = y;
	cvFloodFill( image_, seed, value, lo, up );
}

void Image::fill( double value, int x, int y, int width, int height ) {
	CvPoint left, right;
	left.x = x;
	left.y = y;
	right.x = x + width-1;
	right.y = y + height-1;
	cvRectangle( image_, left, right, value, CV_FILLED );
}
	
void Image::circle( double value, int x, int y, int radius, bool fill ) {
	CvPoint center;
	center.x = x;
	center.y = y;
	cvCircle( image_, center, radius, value, fill? CV_FILLED : 1 );
}

void Image::line( double value, int x1, int y1, int x2, int y2 ) {
	CvPoint p1;
	CvPoint p2;
	p1.x = x1;
	p1.y = y1;
	p2.x = x2;
	p2.y = y2;
	cvLine( image_, p1, p2, value );
}

static CvRect intersection( CvRect r1, CvRect r2 ) {
	int x1 = max( r1.x, r2.x );
	int y1 = max( r1.y, r2.y );
	int x2 = min( r1.x + r1.width,  r2.x + r2.width  );
	int y2 = min( r1.y + r1.height, r2.y + r2.height );
	return cvRect( x1, y1, x2-x1, y2-y1 );
}
	
void Image::blit( const Image& img, int x, int y ) {
	
	// see if fully off screen
	if ( (x+img.width()) <= 0 || (y+img.height()) <= 0 )
		return;
	if ( x >= w_ || y >= h_ )
		return;

	if ( x >= 0 && y >= 0 && (x+img.width()) < w_ && (y+img.height()) < h_ ) {
		CvMat mat;
		CvRect rect = cvRect( x, y, img.width(), img.height() );
		cvGetSubRect( image_, &mat, rect );
		cvCopy( img.image_, &mat );
	}
	else {
		CvRect bgRect = cvRect( 0, 0, w_, h_ );
		CvRect fgRect = cvRect( x, y, img.width(), img.height() );
		CvRect rect = intersection( bgRect, fgRect );
		
		CvMat bgMat;
		cvGetSubRect( image_, &bgMat, rect );
		
		//cout <<"rect.x=" << rect.x << " rect.y=" << rect.y << endl;
		
		bgRect = cvRect( -x, -y, w_, h_ );
		fgRect = cvRect( 0, 0, img.width(), img.height() );
		rect = intersection( bgRect, fgRect );
		
		//cout <<"rect.x=" << rect.x << " rect.y=" << rect.y << endl;
		//cout << endl;
		
		CvMat fgMat;
		cvGetSubRect( img.image_, &fgMat, rect );
		
		cvCopy( &fgMat, &bgMat );
	}
}



void Image::add_noise( double range ) {
	Image img( width(), height() );
	img = 0.0;
	for ( int x = 0; x < w_; x++ ) {
		for ( int y = 0; y < h_; y++ ) {
			img.fill( range*(0.5-randdouble()), x, y, 1, 1 );
		}
	}
	add( *this, img );
}
	
void Image::operator=( const Image& rhs ) {
	if ( this == &rhs )
		return;
	
	if ( rhs.w_ < 0 || rhs.h_ < 0 )
		return;
		
	if ( w_ != rhs.w_ || h_ != rhs.h_ ) {
		w_ = rhs.w_;
		h_ = rhs.h_;
		cvReleaseImage( &image_ );
		CvSize imgSize;
		imgSize.width = w_;
		imgSize.height = h_;
		image_ = cvCreateImage( imgSize, DEPTH, 1);	
	}
	cvCopy(rhs.image_, image_ );
}

void Image::operator=( double rhs ) {
	CvScalar scalar;
	scalar = cvRealScalar( rhs );
	cvSet( image_, scalar );
}

void Image::operator=( unsigned char* rhs ) {
	//cvSetImageData( image_, rhs, w_ );
	unsigned char* pData = NULL;
	int step = w_;
	CvSize size;
	cvGetRawData( image_, &pData, &step, &size );
	if ( pData != NULL ) {
		memcpy( pData, rhs, sizeof( unsigned char ) *w_*h_ );
	}
	else
		throw runtime_error( "Image operator= not working right" );
}

void Image::convert( const Image& image ) {
	cvConvert( image.image_, image_ );
}

void Image::scale( double scale, const Image& image ) {
	cvScale( image.image_, image_, scale );
}

void Image::resize( const Image& image ) {
	cvResize( image.image_, image_ );
}

void Image::add( const Image& a, const Image& b ) {
	cvAdd( a.image_, b.image_, image_ );
}

void Image::sub( const Image& a, const Image& b ) {
	cvSub( a.image_, b.image_, image_ );
}

void Image::mul( const Image& a, const Image& b ) {
	cvMul( a.image_, b.image_, image_ );
}

void Image::div( const Image& a, const Image& b ) {
	cvDiv( a.image_, b.image_, image_ );
}
	
void Image::gt( const Image& a, const Image& b ) {
	Image result( w_, h_, IPL_DEPTH_8U );
	cvCmp( a.image_, b.image_, result.image_, CV_CMP_GT );
	cvConvertScale( result.image_, image_, 1.0/255.0 );
}

void Image::lt( const Image& a, const Image& b ) {
	Image result( w_, h_, IPL_DEPTH_8U );
	cvCmp( a.image_, b.image_, result.image_, CV_CMP_LT );
	cvConvertScale( result.image_, image_, 1.0/255.0 );
}
	
void Image::max( const Image& a, const Image& b ) {
	cvMax( a.image_, b.image_, image_ );
}

void Image::min( const Image& a, const Image& b ) {
	cvMin( a.image_, b.image_, image_ );
}

void Image::absDiff( const Image& a, const Image& b ) {
	cvAbsDiff( a.image_, b.image_, image_ );
}

// image with double
void Image::add( const Image& a, double b ) {
	cvAddS( a.image_, cvRealScalar( b ), image_ );
}

void Image::sub( const Image& a, double b ) {
	cvSubS( a.image_, cvRealScalar( b ), image_ );
}

void Image::mul( const Image& a, double b ) {
	*this = b;
	mul( a, *this );
}

void Image::div( const Image& a, double b ) {
	*this = b;
	div( a, *this );
}

void Image::gt( const Image& a, double b ) {
	Image result( w_, h_, IPL_DEPTH_8U );
	cvCmpS( a.image_, b, result.image_, CV_CMP_GT );
	cvConvertScale( result.image_, image_, 1.0/255.0 );
}

void Image::lt( const Image& a, double b ) {
	Image result( w_, h_, IPL_DEPTH_8U );
	cvCmpS( a.image_, b, result.image_, CV_CMP_LT );
	cvConvertScale( result.image_, image_, 1.0/255.0 );
}

void Image::max( const Image& a, double b ) {
	cvMaxS( a.image_, b, image_ );
}

void Image::min( const Image& a, double b ) {
	cvMinS( a.image_, b, image_ );
}

void Image::absDiff( const Image& a, double b ) {
	cvAbsDiffS( a.image_, image_, cvScalarAll( b ) );
}

// double with image
void Image::add( double a, const Image& b ) {
	cvAddS( b.image_, cvRealScalar( a ), image_ );
}

void Image::sub( double a, const Image& b ) {
	cvSubRS( b.image_, cvRealScalar( a ), image_ );
}

void Image::mul( double a, const Image& b ) {
	*this = a;
	mul( *this, b );
}

void Image::div( double a, const Image& b ) {
	*this = a;
	div( *this, b );
}

void Image::gt( double a, const Image& b ) {
	Image result( w_, h_, IPL_DEPTH_8U );
	cvCmpS( b.image_, a, result.image_, CV_CMP_LE );
	cvConvertScale( result.image_, image_, 1.0/255.0 );
}

void Image::lt( double a, const Image& b ) {
	Image result( w_, h_, IPL_DEPTH_8U );
	cvCmpS( b.image_, a, result.image_, CV_CMP_GE );
	cvConvertScale( result.image_, image_, 1.0/255.0 );
}

void Image::max( double a, const Image& b ) {
	cvMaxS( b.image_, a, image_ );
}

void Image::min( double a, const Image& b ) {
	cvMinS( b.image_, a, image_ );
}

void Image::absDiff( double a, const Image& b ) {
	cvAbsDiffS( b.image_, image_, cvScalarAll( a ) );
}


void Image::gaussian( const Image& image, int w, int h ) {
	cvSmooth( image.image_, image_,CV_GAUSSIAN, w, h );
}

void Image::erode( const Image& image ) {
	cvErode( image.image_, image_ );
}

void Image::dilate( const Image& image ) {
	cvDilate( image.image_, image_ );
}

void Image::canny( const Image& image, double threshold1, double threshold2 ) {
	Image converted( image.width(), image.height(), IPL_DEPTH_8U );
	Image edges( image.width(), image.height(), IPL_DEPTH_8U );
	converted.convert( image );
	
	cvCanny( converted.image_, edges.image_, threshold1, threshold2 );
	convert( edges );
}

void Image::good_features( const Image& image, double qualityLevel, double minDistance ) {
	Image eigenImage( image.width(), image.height(),  IPL_DEPTH_32F );
	Image tempImage(  image.width(), image.height(),  IPL_DEPTH_32F );
	int cornerCount = 20;
  	CvPoint2D32f corners[ cornerCount ];
	
	cvGoodFeaturesToTrack(image.image_, eigenImage.image_, tempImage.image_, corners, &cornerCount, qualityLevel, minDistance );
	
	*this = 0.0;
	for ( int i = 0; i < cornerCount; i++ ) {
		fill( 255.0, (int)corners[ i ].x, (int)corners[ i ].y, 1, 1 );
	}
}

double Image::avg() const {
	CvScalar average;
	average = cvAvg( image_ );
	return average.val[ 0 ];
}

double Image::sum() const {
	CvScalar summation;
	summation = cvSum( image_ );
	return summation.val[ 0 ];
}

void Image::maxLoc( int& x, int& y ) const {
	CvPoint maxPt;
	CvPoint minPt;
	double maxVal;
	double minVal;
	cvMinMaxLoc( image_, &minVal, &maxVal, &minPt, &maxPt );
	x = maxPt.x;
	y = maxPt.y;
}

void Image::centreOfGravity( int& x, int& y ) const {
	CvMoments moments;
	cvMoments( image_, &moments );
	double mom  = cvGetSpatialMoment( &moments, 0, 0 );
	double momx = cvGetSpatialMoment( &moments, 1, 0 );
	double momy = cvGetSpatialMoment( &moments, 0, 1 );
	x = (int)(round( momx/mom ));
	y = (int)(round( momy/mom ));
}

