#include "ImagesEvaluation.h"

#include "Candidate.h"
#include "rand-util.h"

#ifdef SHOW_SIMULATION
#include "opencv/cv.h"
#include "opencv/highgui.h"
#endif

#include <dirent.h>
#include <fstream>

string ImagesEvaluation::IMAGES_DIR( "images" );
	
ImagesEvaluation::ImagesEvaluation( ImageArray& input0, ImageArray& input1, const char* images_dir ) 
 : input0_( input0 ),
   input1_( input1 )
{

	DIR* dir = opendir( images_dir );
	struct dirent* entry = NULL;
	
	while( (entry = readdir( dir )) != NULL ) {
		string name = entry->d_name;
		// uses unsigned int, so can't have -1
		if ( name.rfind( ".jpg" ) == (name.size()-4) ) {
			string pathname = images_dir;
			pathname += "/";
			pathname += name;
			fileNames_.push_back( pathname );
			//cout << pathname << endl;
		}
	}
	
	if ( closedir( dir ) != 0 )
		cerr << "Error closing directory: " << images_dir << endl;
	
	string pathname = images_dir;
	pathname += "/";
	pathname += "coords.txt";
	
	ifstream coordsFile( pathname.c_str() );
	while( !coordsFile.eof() ) {
		vector<int> coords( 8 );
		bool endRead = false;
		for ( int i = 0; i < 8; i++ ) {
			if ( coordsFile.eof() ) {
				endRead = true;
				break;
			}
			coordsFile >> coords[ i ];
			//cout << coords[ i ] << " ";
		}
		//cout << endl;
		if ( !endRead )
			targetCoords_.push_back( coords );
	}
	coordsFile.close();
	
	assert( targetCoords_.size() == fileNames_.size() );
	
#ifdef SHOW_SIMULATION
	cvNamedWindow( "Simulation", 1 );
	cvNamedWindow( "Output", 1 );
#endif
	
	initialise();

}

ImagesEvaluation::~ImagesEvaluation() {
#ifdef SHOW_SIMULATION
	cvDestroyWindow( "Simulation" );
	cvDestroyWindow( "Output" );
#endif
}

void ImagesEvaluation::initialise( int numcases ) {
	assert( (unsigned int)numcases <= fileNames_.size() );
			
	testCases_.clear();
	
	for ( int i = 0; i < numcases; i++ ) {
		int index = randrange( fileNames_.size() );
		string filename = fileNames_[ index ];
		vector<int> coords = targetCoords_[ index ];
		ImageTestCase* testCase = new ImageTestCase;
		testCases_.add( testCase );
		testCase->image.load( filename.c_str() );
		assert( coords.size() == 8 );
		for ( int i = 0; i < 8; i+=2 ) {
			CvPoint& pt = testCase->target[ i/2 ];
			pt.x = coords[ i ];
			pt.y = coords[ i+1 ];
		}
	}
}

double ImagesEvaluation::evaluate( Candidate<ImageArray>& candidate ) {
	double worst = 0.0;
	ImageArray result;
	for ( unsigned int i = 0; i < testCases_.size(); i++ ) {
		ImageTestCase& testCase = testCases_[ i ];
		input0_.image() = testCase.image;
		input1_ = input0_;
		
		
		candidate.program().evaluate( result );
		const Image& result_img = result.image();
		
		int x = 0, y = 0;
		result_img.maxLoc( x, y );

		
#ifdef SHOW_SIMULATION
		if ( randdouble() < 0.005) {
			Image& image = input0_.image();
			for ( int i = 0; i < 4; i++ ) {
				CvPoint pt1 = testCase.target[ i ];
				CvPoint pt2 = testCase.target[ (i+1)%4 ];
				image.line( 255.0, pt1.x, pt1.y, pt2.x, pt2.y );
			}
			image.line( 255.0, 0, y, image.width(), y );
			image.line( 255.0, x, 0, x, image.height() );
			cvShowImage("Simulation",image.image() );
			cvShowImage( "Output", result_img.image() );
			cvWaitKey(5);
		}
#endif
		
		//CvRect r = testCase.target;
		//if ( r.x >= x && x <= (r.x+r.width) && r.y >= y && y <= (r.y+r.height) )
		//	sum++;
		int avx = 0, avy = 0;
		for ( int i = 0; i < 4; i++ ) {
			CvPoint pt1 = testCase.target[ i ];
			avx += pt1.x;
			avy += pt1.y;
		}
		avx /= 4;
		avy /= 4;
		int dx = avx-x, dy = avy-y;
		double distSq = (dx*dx) + (dy*dy);
		worst = max( worst, distSq );
	}
	//cout << testCases_.size() << endl;
	return worst;
}


