#include "MotionMinimalSim.h"

#include "opencv/cv.h"
#include "opencv/highgui.h"

#include "rand-util.h"
#include "Properties.h"

#include <string>
#include <fstream>
using namespace std;

bool MotionMinimalSim::PURE_MOTION = false;

MotionMinimalSim::MotionMinimalSim()
 : background_( ImageArray::WIDTH, ImageArray::HEIGHT ),
   object_( 0, 0 ),
   buffer_( ImageArray::WIDTH, ImageArray::HEIGHT ),
   positions_( 3 ),
   noise_( 3 ),
   epoch_( 0 ),
   realData_( false )
{
	//initialise();
#ifdef SHOW_SIMULATION
	cvNamedWindow( "Motion", 1 );
	cvNamedWindow( "Static", 1 );
#endif
}

MotionMinimalSim::~MotionMinimalSim() {
#ifdef SHOW_SIMULATION
	cvDestroyWindow( "Motion" );
	cvDestroyWindow( "Static" );
#endif
}
	
void MotionMinimalSim::initialise() {
	//cout << "initialise: begin" << endl;
	realData_ = false;
	int w = ImageArray::WIDTH/20;
	int objWidth  = w + randrange( w );
	int objHeight = w + randrange( w );
	Image object( objWidth, objHeight );
	
	
	if ( !PURE_MOTION ) {
		double colorRange = 1.0;
	
		double fg = 1.0;
		double bg = 0.0;
		do {
			fg = 1.0 - randdouble()*colorRange;
			fg = max( 0.0, min( 1.0, fg ) );
			bg = randdouble()*colorRange;
			bg = max( 0.0, min( 1.0, bg ) );
		}
		while( abs( bg-fg ) <= NOISE );
		object = (double)fg;
		object.add_noise( NOISE );
		object_ = object;
		background_ = (double)bg;
		background_.add_noise( NOISE );
		
		
		
	}
	else {
		object = 0.5;
		object.add_noise( 1.0 );
		object_ = object;
		background_ = 0.5;
		background_.add_noise( 1.0 );
		
	}
	
	// make it start a bit away from the edges
	for ( unsigned int i = 0; i < positions_.size(); i++ ) {
		vector2d& pos = positions_[ i ];
		pos.x = object_.width()  + randrange( background_.width()  - 2*object_.width()  );
		pos.y = object_.height() + randrange( background_.height() - 2*object_.height() );
	}
	
	
	for ( unsigned int i = 0; i < noise_.size(); i++ ) {
		Image noise( ImageArray::WIDTH, ImageArray::HEIGHT );
		noise = 0.0;
		noise.add_noise( NOISE );
		noise_[ i ] = noise;
		//cout << noise_[ i ].width()  << " " << noise_[ i ].height() << endl;
	}
	
	epoch_++;
	//cout << "Epoch=" << epoch_ << endl;
	//cout << "initialise: done" << endl;
}

void MotionMinimalSim::initialise( const char* file ) {
	realData_ = true;

	ifstream infile( file );
	Properties properties( infile );
	infile.close();
	
	vector<string> imageNames = properties.keys();
	
	vector<vector2d> positions;
	vector<Image>    images;
	
	for ( unsigned int i = 0; i < imageNames.size(); i++ ) {
		string imageName = imageNames[ i ];
		Image image( imageName.c_str() );
		
		
		string posStr = properties.get( imageName.c_str(), "" );
		istringstream instr( posStr );
		vector2d position;
		instr >> position;
		
		images.push_back( image );
		positions.push_back( position );
	}
	
	images_ = images;
	positions_ = positions;
}

double MotionMinimalSim::evaluate( Candidate<ImageArray>& candidate ) {
	if ( !PURE_MOTION )
		return 0.5*evaluateFunction0( candidate ) + 0.5*evaluateFunction1( candidate );
	else
		return 0.5*evaluateFunction0( candidate ) + 0.5*evaluateFunction1PureMotion( candidate );
	//return evaluateFunction1PureMotion( candidate );
	//return evaluateFunction0( candidate );
	//return evaluateFunction1( candidate );
}

double MotionMinimalSim::evaluateFunction0( Candidate<ImageArray>& candidate ) {

	GPProgram<ImageArray>& program = candidate.program();
	vector<ImageArray*> inputs = program.tree( 0 ).inputs();
	ImageArray& input0 = *inputs[ 0 ];
	ImageArray& input1 = *inputs[ 1 ];

	int duration = positions_.size()-1;

#ifdef SHOW_SIMULATION
	bool showSim = randdouble() < 0.01;
	vector2d tracking = positions_[ 0 ];
#endif

	double errorSq = 0.0;
	//double maxDistSq = buffer_.width()*buffer_.width() + buffer_.height()*buffer_.height();

	vector2d objectPos = positions_[ 0 ];
	vector2d target( objectPos );

	ImageArray output;

	// we do one extra step at start to get things going
	// so input0 will contain frame from last time
	int steps = 0;
	for ( int i = -1; i < duration; i++, steps++ ) {
		
		objectPos = positions_[ i+1 ];
		
		input0 = buffer_;
		
		if ( !realData_ ) {
			buffer_ = background_;
			buffer_.blit( object_, (int)(objectPos.x-object_.width()/2), (int)(objectPos.y-object_.height()/2) );
			buffer_.add( buffer_, noise_[ i+1 ] );
		}
		else {
			buffer_ = images_[ i+1 ];
		}
			
		input1 = buffer_;
		
		if ( i >= 0 ) {
			
			//input0 = positions_[ i ];
			//input1 = positions_[ i+1 ];
			
			program.evaluate( output, 0 );
			
			int x = 0, y = 0;
			vector2d vel = output.toVector();
			target = positions_[ i ] + vel;
			x = (int)target.x;
			y = (int)target.y;
			// compute difference from velocity
			double dx = target.x-objectPos.x;
			double dy = target.y-objectPos.y;
			double diffSq = dx*dx + dy*dy;
			errorSq += diffSq;
			//double errSq = (diffSq/maxDistSq);
			//errorSq += errSq;
			
#ifdef SHOW_SIMULATION
			if ( showSim ) {
				tracking = tracking + vel;
				Image& image = input1.image();
				x = (int)tracking.x;
				y = (int)tracking.y;
				//image.line( 255.0, object_x, object_y, (int)(object_x + velocity_[ 0 ]), (int)(object_y + velocity_[ 1 ]) );
				//image.line( 255.0, object_x, object_y, (int)(object_x + vel.x), (int)(object_y + vel.y) );
				image.line( 255.0, 0, y, image.width(), y );
				image.line( 255.0, x, 0, x, image.height() );
				// draw box
				/*image.line( 255.0, object_x,                   object_y, 
				                   object_x + object_.width(), object_y );
				image.line( 255.0, object_x,                   object_y,
				                   object_x,                   object_y + object_.height() );
				image.line( 255.0, object_x,                   object_y + object_.height(),
				                   object_x + object_.width(), object_y + object_.height() );
				image.line( 255.0, object_x + object_.width(), object_y,
				                   object_x + object_.width(), object_y + object_.height() );*/
				cvShowImage("Motion",image.image() );
				cvWaitKey(40);
			}
#endif
	
		}
	}
	
	// should be between 0 and 1
	return errorSq/steps;
}

double MotionMinimalSim::evaluateFunction1( Candidate<ImageArray>& candidate ) {
	GPProgram<ImageArray>& program = candidate.program();
	vector<ImageArray*> inputs = program.tree( 1 ).inputs();
	ImageArray& input0 = *inputs[ 0 ];

	//double maxDistSq = buffer_.width()*buffer_.width() + buffer_.height()*buffer_.height();

	double diffSq = 0.0;
	ImageArray output;

	int duration = realData_? positions_.size() : 1; // only test one image when not using real data

	for ( int i = 0; i < duration; i++ ) {

		vector2d objectPos = positions_[ i ];
		
		if ( !realData_ ) {
			buffer_ = background_;
			// draw centered
			buffer_.blit( object_, (int)(objectPos.x-object_.width()/2), (int)(objectPos.y-object_.height()/2) );
			buffer_.add( buffer_, noise_[ i ] );
		}
		else {
			buffer_ = images_[ i ];
		}
		
		input0 = buffer_;
		
		program.evaluate( output, 1 );
			
		vector2d pos = output.toVector();

#ifdef SHOW_SIMULATION
		if ( randdouble() < 0.01 ) {
			Image& image = input0.image();
			int x = (int)pos.x;
			int y = (int)pos.y;
			image.line( 255.0, 0, y, image.width(), y );
			image.line( 255.0, x, 0, x, image.height() );
			cvShowImage("Static", image.image() );
			cvWaitKey(40);
		}
#endif
			
		double dx = pos.x-objectPos.x;
		double dy = pos.y-objectPos.y;
		diffSq += dx*dx + dy*dy;
	}
	return diffSq/duration;
}

double MotionMinimalSim::evaluateFunction1PureMotion( Candidate<ImageArray>& candidate ) {

	GPProgram<ImageArray>& program = candidate.program();
	vector<ImageArray*> inputs = program.tree( 0 ).inputs();
	ImageArray& input0 = *inputs[ 0 ];
	ImageArray& input1 = *inputs[ 1 ];

	int duration = positions_.size()-1;

	double errorSq = 0.0;
	//double maxDistSq = buffer_.width()*buffer_.width() + buffer_.height()*buffer_.height();

	vector2d objectPos = positions_[ 0 ];

	ImageArray output;

	// we do one extra step at start to get things going
	// so input0 will contain frame from last time
	int steps = 0;
	for ( int i = -1; i < duration; i++, steps++ ) {
		
		objectPos = positions_[ i+1 ];
		
		input0 = buffer_;
		
		buffer_ = background_;
		buffer_.blit( object_, (int)(objectPos.x-object_.width()/2), (int)(objectPos.y-object_.height()/2) );
		buffer_.add( buffer_, noise_[ i+1 ] );
			
		input1 = buffer_;
		
		if ( i >= 0 ) {
			
			program.evaluate( output, 1 );
			
			vector2d pos = output.toVector();
			
			double diffSq = (objectPos - pos).dotProduct();
			errorSq += diffSq;
		}
	}
	
	// should be between 0 and 1
	return errorSq/steps;
}
