#ifndef __SteadyStateGA_H__
#define __SteadyStateGA_H__

#include <limits>
#include <cfloat>
using namespace std;

#include "GPTree.h"
#include "Evaluation.h"
#include "Candidate.h"
#include "rand-util.h"
#include "auto_list.h"

template<class T> class SteadyStateGA {

public:
	SteadyStateGA( Evaluation<T>& evaluation, int pop_size, auto_list< ADFunction<T> >& definedFunctions );
	virtual ~SteadyStateGA();
	
public:
	double best_fitness() { return best_.fitness(); };
	int best_program_size() { return best_.program_size(); };
	const Candidate<T>& best() const { return best_; };
	
	double average_fitness();
	double average_program_size();
	
	void do_step();
	
	void force_evaluations();
	void recalculate_best();

protected:
	Candidate<T>& tournament();
	void random_replace( Candidate<T>& child );
	void retain_best( Candidate<T>& candidate );
	int  duplicates( const Candidate<T>& candidate ) const; 
	bool is_better( Candidate<T>& candidate1, Candidate<T>& candidate2 );

public:
	static double RECOMBINATION;
	static double MUTATION;
	static double REPLACE;
	
	static double EXPECTED_ERROR;
	static int EXPECTED_COMPLEXITY;
	
	static int    TOURNAMENT_SIZE;
	static int    POP_SIZE;
	static int    ITERATIONS;
		
protected:
	Candidate<T> best_;

	Candidate<T> child_;

	Candidate<T>* population_;
	int pop_size_;
	

};

template<class T> double SteadyStateGA<T>::RECOMBINATION   = 0.9;
template<class T> double SteadyStateGA<T>::MUTATION        = 0.01;
template<class T> double SteadyStateGA<T>::REPLACE         = 0.002;
template<class T> double SteadyStateGA<T>::EXPECTED_ERROR  = 0.0;
template<class T> int    SteadyStateGA<T>::TOURNAMENT_SIZE = 2;
template<class T> int    SteadyStateGA<T>::EXPECTED_COMPLEXITY = 20;
template<class T> int    SteadyStateGA<T>::POP_SIZE        = 1000;
template<class T> int    SteadyStateGA<T>::ITERATIONS      = 10000;


template<class T> SteadyStateGA<T>::SteadyStateGA( Evaluation<T>& evaluation, int pop_size, auto_list< ADFunction<T> >& definedFunctions )
 : best_( evaluation, definedFunctions, true ),
   child_( evaluation, definedFunctions, false ),
   population_( NULL ),
   pop_size_( pop_size )
{
	// initialise is true for these
	population_ = new Candidate<T>[ pop_size_ ]( evaluation, definedFunctions, true );
}

template<class T> SteadyStateGA<T>::~SteadyStateGA() {
	if ( population_ != NULL ) {
		delete[] population_;
		population_ = NULL;
	}
}

template<class T> Candidate<T>& SteadyStateGA<T>::tournament() {
	assert( TOURNAMENT_SIZE > 0 );
	Candidate<T>* best = NULL;
	for ( int i = 0; i < TOURNAMENT_SIZE; i++ ) {
		Candidate<T>& candidate = population_[ randrange( pop_size_ ) ];
		if ( best == NULL || is_better( candidate, *best ) )
			best = &candidate;
	}
	
	return *best;
}

template<class T> void SteadyStateGA<T>::random_replace( Candidate<T>& child ) {
	Candidate<T>& replace = population_[ randrange( pop_size_ ) ];
	if ( is_better( child, replace ) || randdouble() < REPLACE )
		replace = child;
}

template<class T> void SteadyStateGA<T>::retain_best( Candidate<T>& candidate ) {
	/*if ( candidate.fitness() < best_.fitness() 
	||  (candidate.fitness() == best_.fitness() 
	  && candidate.program_size() < best_.program_size() ) )
	{*/
	if ( is_better( candidate, best_ ) ) {
		best_ = candidate;
	}
}

template<class T> int SteadyStateGA<T>::duplicates( const Candidate<T>& candidate ) const {
	int num = 0;
	for ( int i = 0; i < pop_size_; i++ ) {
		if ( population_[ i ] == candidate )
			num++;
	}
	return num;
}

template<class T> bool SteadyStateGA<T>::is_better( Candidate<T>& candidate1, Candidate<T>& candidate2 ) {
	double fitness1 = candidate1.fitness();
	double fitness2 = candidate2.fitness();
	
	if ( (fitness1 > EXPECTED_ERROR || fitness2 > EXPECTED_ERROR) /*&& fitness1 != fitness2*/ )
		return fitness1 < fitness2;
	
	return candidate1.program_size() < candidate2.program_size();
}

template<class T> double SteadyStateGA<T>::average_fitness() {
	double sum = 0;
	for ( int i = 0; i < pop_size_; i++ ) {
		sum += population_[ i ].fitness();
	}
	return sum/pop_size_;
}

template<class T> double SteadyStateGA<T>::average_program_size() {
	int sum = 0;
	for ( int i = 0; i < pop_size_; i++ ) {
		sum += population_[ i ].program().size();
	}
	return sum/(double)pop_size_;
}

template<class T> void SteadyStateGA<T>::do_step() {
	Candidate<T>& parent1 = tournament();
	Candidate<T>& parent2 = tournament();
	
	child_ = parent1;

	if ( randdouble() < RECOMBINATION )
		child_.recombine( parent2 );
	
	if ( randdouble() < MUTATION )
		child_.mutate();
	
	retain_best( child_ );
		
	random_replace( child_ );
}

template<class T> void SteadyStateGA<T>::force_evaluations() {
	for ( int i = 0; i < pop_size_; i++ ) {
		population_[ i ].force_evaluation();
	}
	best_.force_evaluation();
}


template<class T> void SteadyStateGA<T>::recalculate_best() {
	for ( int i = 0; i < pop_size_; i++ ) {
		retain_best( population_[ i ] );
	}
}

#endif // __SteadyStateGA_H__
