#include "TestInt.h"

#include <time.h>
#include <sys/time.h>
#include <iostream>
#include <vector>
#include <limits>
using namespace std;

#include "GPNode.h"
#include "Function.h"
#include "Constant.h"
#include "Variable.h"
#include "BinaryOperator.h"
#include "GPTree.h"
#include "SteadyStateGA.h"
#include "Evaluation.h"
#include "Candidate.h"
#include "ADFunction.h"
#include "auto_list.h"

void add_op( const int& a, const int& b, int& output ) {
	output = a+b;
}

void sub_op( const int& a, const int& b, int& output ) {
	output = a-b;
}

void mul_op( const int& a, const int& b, int& output ) {
	output = a*b;
}

void div_op( const int& a, const int& b, int& output ) {
	if ( a != 0 && b != 0 )
		output = a/b;
	else
		output = 1;
}

unsigned int evaluate( int& x, int& y, GPTree<int>& tree ) {
	unsigned int sum = 0;
	for ( x = 0; x < 10; x++ ) {
		for ( y = 0; y < 10; y++ ) {
			unsigned int z1 = x*y + 2*x + x*x + 10;
			int z2;
			tree.evaluate( z2 );
			unsigned int diff = z1-z2;
			sum += diff*diff;
		}
	}
	return sum;
}

class TestEval : public Evaluation<int> {
	
public:
	TestEval( int& x, int& y ) : x_( &x ), y_( &y ) { };
	~TestEval() {};

public:
	void initialise( int numcases = TEST_CASES ) {}

	double evaluate( Candidate<int>& candidate ) {
		GPProgram<int>& program = candidate.program();
		int& x = *x_;
		int& y = *y_;
		unsigned int sum = 0;
		for ( x = 0; x < 10; x++ ) {
			for ( y = 0; y < 10; y++ ) {
				unsigned int z1 = 3*x*y + 2*x + x*x + y*y + 10;
				int z2;
				program.evaluate( z2 );
				unsigned int diff = z1-z2;
				sum += diff*diff;
			}
		}
		return (double)sum;
	};
	
protected:
	int* x_;
	int* y_;
};

static void setup( int& x, int& y, auto_list< Function<int> >& terminals, auto_list< Function<int> >& functions ) {
	terminals.add( new Constant<int>( "1", 1 ) );
	terminals.add( new Constant<int>( "2", 2 ) );
	terminals.add( new Constant<int>( "3", 3 ) );
	terminals.add( new Constant<int>( "4", 4 ) );
	terminals.add( new Constant<int>( "5", 5 ) );
	terminals.add( new Constant<int>( "6", 6 ) );
	terminals.add( new Constant<int>( "7", 7 ) );
	terminals.add( new Constant<int>( "8", 8 ) );
	terminals.add( new Constant<int>( "9", 9 ) );
	terminals.add( new Constant<int>( "10", 10 ) );

	terminals.add( new Variable<int>( "x", x ) );
	terminals.add( new Variable<int>( "y", y ) );

	functions.add( new BinaryOperator<int>( "+", add_op ) );
	functions.add( new BinaryOperator<int>( "-", sub_op ) );
	functions.add( new BinaryOperator<int>( "*", mul_op ) );
	functions.add( new BinaryOperator<int>( "/", div_op ) );
}

void testint () {
	struct timeval tv;
    if ( gettimeofday( &tv, NULL ) == 0 ) {
        srand((unsigned)(tv.tv_sec + tv.tv_usec));
    }
    else {
        srand( time( NULL ) );
    }

	int x = 0;
	int y = 0;

	auto_list< Function<int> > terminals;
	auto_list< Function<int> > functions;
	
	setup( x, y, terminals, functions );

	GPProgram<int>::ADFS.push_back( 1 );
	for ( unsigned int i = 0; i < GPProgram<int>::ADFS.size(); i++ ) {
		ostringstream strout;
		strout << "fn" << i;
		ADFunction<int>* adf = new ADFunction<int>( strout.str().c_str(), GPProgram<int>::ADFS[ i ] );
		if ( adf->arity() == 0 )
			terminals.add( adf );
		else
			functions.add( adf );
		
		setup( x, y, adf->terminals(), adf->functions() );
	}
    
	TestEval eval( x, y );
	
	int pop_size = 1000;

	
	SteadyStateGA<int> ga( eval, pop_size, terminals, functions );
	
	for ( int i = 0; i < 100000; i++ ) {
		ga.do_step();
		if ( i % pop_size == 0 ) {
			cout << i << endl;
			cout << "Fitness=" << ga.best_fitness()   << endl;
			cout << "Size="    << ga.best_program_size() << endl;
			cout << "Average Fitness=" << ga.average_fitness() << endl;
			cout << "Average Size=" << ga.average_program_size() << endl;
		}
	}
	
	cout << "Fitness=" << ga.best_fitness()   << endl;
	cout << "Size="    << ga.best_program_size() << endl;
	cout << "Average Fitness=" << ga.average_fitness() << endl;
	cout << "Average Size=" << ga.average_program_size() << endl;
	cout << ga.best()                  << endl;
	
}
